/*
 * The Kibitz Indexer
 *
 * (c) 2005-2014 Tomas Valla <tom@ucw.cz>
 *
 * This software may be freely distributed and used according to the terms
 * of the GNU Lesser General Public License. See COPYING for details.
 */

#undef LOCAL_DEBUG

#include <kibitz/lib.h>
#include <ucw/conf.h>
#include <ucw/fastbuf.h>
#include <ucw/ff-binary.h>
#include <ucw/heap.h>
#include <ucw/mempool.h>
#include <ucw/slists.h>
#include <ucw/stkstring.h>
#include <ucw/unicode.h>
#include <kibitz/index.h>
#include <kibitz/internal.h>
#include <custom/custom.h>
#include <fcntl.h>
#include <time.h>

#define TRACE(x...) do { if (config.indexer_verbose) msg(L_INFO, x); } while(0)

static struct fastbuf *idx;
static struct index_header index_header;

static void write_tmp_header(void)
{
	struct index_header hdr;

	DBG("Writing temp header...");
	bzero(&hdr, sizeof(hdr));
	hdr.magic = INDEX_TEMP_MAGIC;
	hdr.version = INDEX_FORMAT_VERSION;
	hdr.timestamp = time(NULL);

	bwrite(idx, &hdr, sizeof(hdr));
	DBG("Done.");
}

struct phrase_data {
	snode sn;
	u32 rel_offset;
	struct phrase *phrase;
	int inside_ff_heap;
	int quality;		// used by fflist builder
};

static slist phrase_list;
static struct mempool *phrase_list_mp;
static uns phrase_cnt;

static void store_phrasebuf(void)
{
	snode *sn;
	uns padsize = (~(uns)btell(idx) + 1) & 3;

	TRACE("Storing phrase buf...");
	DBG("Padding %d bytes", padsize);
	while (padsize--)
		bputc(idx, 0);
	index_header.phrase_start = btell(idx);

	SLIST_WALK(sn, phrase_list) {
		struct phrase_data *pd = (typeof(pd)) sn;
		pd->rel_offset = btell(idx) - index_header.phrase_start;
		DBG("phrase='%s' rel_off=%d", pd->phrase->text, pd->rel_offset);
		bwrite(idx, pd->phrase, sizeof(struct phrase) + pd->phrase->text_len);

		padsize = (~(uns)btell(idx) + 1) & 3;
		DBG("Padding %d bytes", padsize);
		while (padsize--)
			bputc(idx, 0);
	}
	TRACE("Phrase buffer stored.");
}

struct word_ref {
	struct phrase_data *pd;
	u16 wstart;
	u16 wlen;
};

static uns wordref_cnt;
static struct word_ref *wordref_buf;
static uns wordref_buf_size;

static inline int wordref_lt(struct word_ref a, struct word_ref b)
{
	int r = strncmp(a.pd->phrase->text+a.wstart, b.pd->phrase->text+b.wstart, MIN(a.wlen, b.wlen));
	if (r)
		return r<0;
	COMPARE_LT(a.wlen, b.wlen);
	return 0;
}

#define ASORT_PREFIX(x) wordref_##x
#define ASORT_KEY_TYPE struct word_ref
#define ASORT_ELT(i) wordref_buf[i]
#define ASORT_LT(x,y) wordref_lt(x,y)
#include <ucw/sorter/array-simple.h>

static struct phrase_data *cur_pd;

void wordref_add(uns wstart, uns wlen)
{
	byte *x = stk_strndup(cur_pd->phrase->text + wstart, wlen);
	uns l = utf8_strlen(x);
	if (l < config.min_word_len || l > config.max_word_len)
		return;
	DBG("Adding word '%s' len=%d", cur_pd->phrase->text+wstart, wlen);
	wordref_cnt++;
	if (unlikely(wordref_cnt > wordref_buf_size)) {
		msg(L_WARN, "Reallocating wordref_buf, you should check the buffer size config");
		wordref_buf_size *= 2;
		wordref_buf = xrealloc(wordref_buf, wordref_buf_size * sizeof(*wordref_buf));
	}
	uns w = wordref_cnt - 1;
	wordref_buf[w].pd = cur_pd;
	wordref_buf[w].wstart = wstart;
	wordref_buf[w].wlen = wlen;
}

static void build_wordrefbuf(void)
{
	TRACE("Building wordref buffer...");
	wordref_buf = xmalloc(wordref_buf_size * sizeof(*wordref_buf));
	SLIST_WALK(cur_pd, phrase_list) {
		DBG("Parsing phrase '%s'", cur_pd->phrase->text);
		custom_get_words(cur_pd->phrase);
	}

	wordref_sort(wordref_cnt);
	TRACE("Wordref buffer built, %d individual words.", wordref_cnt);
}

static void load_phrases(struct fastbuf *input)
{
	TRACE("Loading phrases...");
	slist_init(&phrase_list);
	struct phrase *p;
	while (p = custom_get_phrase(input, phrase_list_mp)) {
		uns l = utf8_strlen(p->text);
		if (l < config.min_phrase_len || l > config.max_phrase_len)
			continue;
		DBG("Got input phrase='%s' relev=%d", p->text, p->relevance);
		struct phrase_data *pd = mp_alloc_zero(phrase_list_mp, sizeof(*pd));
		pd->phrase = p;
		slist_add_tail(&phrase_list, &pd->sn);
		phrase_cnt++;
	}
	index_header.num_phrases = phrase_cnt;
	TRACE("Phrases loaded, %d phrases.", phrase_cnt);
}

static void store_wordoffbuf(void)
{
	TRACE("Storing wordoff buffer...");
	uns pad = (~(uns)btell(idx) + 1) & 3;
	while (pad--)
		bputc(idx, 0);

	index_header.num_wordoffs = wordref_cnt;
	index_header.wordoff_start = btell(idx);

	uns i;
	for (i=0; i<wordref_cnt; i++) {
		DBG("phrase offset=0x%x for word='%s' wstart=%d phrase='%s'", wordref_buf[i].pd->rel_offset,
			wordref_buf[i].pd->phrase->text + wordref_buf[i].wstart,
			wordref_buf[i].wstart,
			wordref_buf[i].pd->phrase->text);
		bputl(idx, wordref_buf[i].pd->rel_offset);
	}
	TRACE("Wordoff buffer stored.");
}

struct trie_node {
	u32 wref;		// index into wordref buffer
	u16 match_size;		// number of character from word matches
	struct trie_node *son, *brother;

	/* Postprocessing data */
	union {
		u32 ff_offset;
		u32 match_cnt;		/* Meaningful only for leaves */
	};
	u32 exact_match;
	u8 ff_cnt;
	u8 son_cnt;
	u32 subtree_size;
};

static struct trie_node *trie_root;
static struct mempool *trie_mempool;

static void trie_init(void)
{
	trie_root = NULL;
        trie_mempool = mp_new(1024*1024);
}

static void trie_cleanup(void)
{
        mp_delete(trie_mempool);
}

/* Inserting _must_ be called on unique words in lexically ascending order */
static void trie_insert(uns wref, uns exact_match)
{
	struct trie_node *t = trie_root;
	byte *wdata = wordref_buf[wref].pd->phrase->text + wordref_buf[wref].wstart;
	uns wlen = wordref_buf[wref].wlen;
	uns i;

	DBG("Inserting wref=%d wstart=%d wlen=%d wdata='%s' exact_match=%d",
			wref, wordref_buf[wref].wstart, wlen, wdata, exact_match);
	ASSERT(exact_match);

	if (!t) {
		DBG("Creating new trie root");
		trie_root = mp_alloc_zero(trie_mempool, sizeof(*trie_root));
		trie_root->wref = wref;
		trie_root->match_size = wlen;
		trie_root->exact_match = exact_match;
		return;
	}

	for (;;) {
		byte *word = wordref_buf[t->wref].pd->phrase->text + wordref_buf[t->wref].wstart;

		DBG("Now in vertex with wref=%d word='%s' match_size=%d", t->wref, word, t->match_size);

		for (i=0; i<MIN(wlen, t->match_size); i++)
			if (word[i] != wdata[i])
				break;

		if (i == MIN(wlen, t->match_size)) {

			DBG("Words matched fully");
			if (!t->son) {
				DBG("Making new son");
				t->son = mp_alloc_zero(trie_mempool, sizeof(*t->son));
				t->son->wref = wref;
				t->son->match_size = wlen;
				t->son->exact_match = exact_match;
				return ;
			}

			struct word_ref wr = wordref_buf[t->son->wref];
			byte sonchar = wr.pd->phrase->text[wr.wstart + i];

			if (sonchar == wdata[i]) {
				DBG("Descending into the son");
				t = t->son;
				continue;
			}

			DBG("Adding new brother of son");
			struct trie_node *oldson = t->son;
			t->son = mp_alloc_zero(trie_mempool, sizeof(*t->son));
			t->son->brother = oldson;
			t->son->wref = wref;
			t->son->match_size = wlen;
			t->son->exact_match = exact_match;
			return;
		}

		struct trie_node *nt = mp_alloc_zero(trie_mempool, sizeof(*nt));
		*nt = *t;
		t->match_size = i;
		nt->brother = NULL;
		DBG("Splitting current vertex into father(wref=%d, match_size=%d)", t->wref, i);

		t->son = mp_alloc_zero(trie_mempool, sizeof(*t->son));
		t->son->match_size = wlen;
		t->son->brother = nt;
		t->son->wref = wref;
		t->son->exact_match = exact_match;
		DBG("... and new son(wref=%d, match_size=%d)", t->son->wref, t->son->match_size);
		return;
	}
}

#ifdef LOCAL_DEBUG
static void trie_dump(struct trie_node *t, uns lev)
{
	char buf[2*lev+1];
	uns i;

	if (!t)
		return;

	for (i=0; i<2*lev; i++)
		buf[i] = ' ';
	buf[2*lev] = 0;

	char word[t->match_size+1];
	strncpy(word, wordref_buf[t->wref].pd->phrase->text + wordref_buf[t->wref].wstart, t->match_size);
	word[t->match_size] = 0;

	DUMP("%s%s=%p wref=%d [%s] ('%s')\n"
	     "%smatch_size=%d match_cnt=%d exact_match=%d son_cnt=%d subtree_size=%d ff_cnt=%d son=%p brother=%p\n",
		buf, (t->match_size == wordref_buf[t->wref].wlen ? "NODE" : "node"), t, t->wref, word,
		wordref_buf[t->wref].pd->phrase->text + wordref_buf[t->wref].wstart, buf, t->match_size,
		t->match_cnt, t->exact_match, t->son_cnt, t->subtree_size, t->ff_cnt, t->son, t->brother);

	t = t->son;
	while (t) {
		trie_dump(t, lev+1);
		t = t->brother;
	}
}
#endif

static uns trie_compute_matches(struct trie_node *t)
{
	ASSERT(t);

	uns m = 0;
	struct trie_node *s;
	for (s = t->son; s; s = s->brother)
		m += trie_compute_matches(s);
	if (t->match_size == wordref_buf[t->wref].wlen)	// We have match at the internal node
		m += t->exact_match;
	else
		t->exact_match = 0;
	t->match_cnt = m;
	ASSERT(m);
	return m;
}

#define HEAP_LESS(x,y) (wordref_buf[x].pd->quality < wordref_buf[y].pd->quality)
static uns store_fflist(struct trie_node *t)
{
	ASSERT(t);
	u32 heap[HARD_FASTFIND_ITEMS+1];
	uns i, j, r = 0;
	u32 off;

	ASSERT(t->match_cnt);
	off = index_header.num_fastfinds;
	for (i=0, j=0; i < config.fastfind_items+1 && j < t->match_cnt; j++) {
		uns wr = t->wref + j;
		struct phrase_data *pd = wordref_buf[wr].pd;
		int q = pd->phrase->relevance;
		if (wr < t->wref + t->exact_match)
			q *= config.exact_match_bonus;

		if (unlikely(pd->inside_ff_heap)) {
			for (uns k=1; k<=i; k++)	// increase the previous occurence quality
				if (wordref_buf[heap[k]].pd == pd) {
					pd->quality += q;
					break;
				}
		}
		else {
			heap[++i] = wr;
			pd->inside_ff_heap = 1;
			pd->quality = q;
		}
	}
	HEAP_INIT(u32, heap, i, HEAP_LESS, HEAP_SWAP);

	for (; j < t->match_cnt; j++) {
		uns wr = t->wref + j;
		struct phrase_data *pd = wordref_buf[wr].pd;
		int q = pd->phrase->relevance;
		if (wr < t->wref + t->exact_match)
			q *= config.exact_match_bonus;

		if (unlikely(pd->inside_ff_heap)) {
			for (uns k=1; k<=i; k++)
				if (wordref_buf[heap[k]].pd == pd) {
					HEAP_INCREASE(u32, heap, i, HEAP_LESS, HEAP_SWAP, k, pd->quality + q);
					break;
				}
		}
		else {
			wordref_buf[heap[1]].pd->inside_ff_heap = 0;
			HEAP_DELETE_MIN(u32, heap, i, HEAP_LESS, HEAP_SWAP);
			pd->inside_ff_heap = 1;
			pd->quality = q;
			HEAP_INSERT(u32, heap, i, HEAP_LESS, HEAP_SWAP, wr);
		}
	}

	if (i == config.fastfind_items+1) {
		wordref_buf[heap[1]].pd->inside_ff_heap = 0;
		HEAP_DELETE_MIN(u32, heap, i, HEAP_LESS, HEAP_SWAP);
	}
	r = i;
	u32 ffbuf[2*r];
	while (i) {
		struct phrase_data *pd = wordref_buf[heap[1]].pd;
		ffbuf[2*(i-1)] = pd->rel_offset;
		ffbuf[2*(i-1) + 1] = pd->quality;
		pd->inside_ff_heap = 0;
		HEAP_DELETE_MIN(u32, heap, i, HEAP_LESS, HEAP_SWAP);
	}
	bwrite(idx, ffbuf, sizeof(ffbuf));
	ASSERT(r);
	t->ff_cnt = r;
	t->ff_offset = off;

	return r;
}

static void build_fflist(struct trie_node *t)
{
	ASSERT(t);

	if (t->match_cnt < config.fastfind_threshold)
		return;
	store_fflist(t);
	index_header.num_fastfind_vertices++;
	index_header.num_fastfinds += t->ff_cnt;
	for (t = t->son; t; t = t->brother)
		build_fflist(t);
}

static void store_fastfindbuf(void)
{
	TRACE("Storing fastfind buffer...");
	uns pad = (~(uns)btell(idx) + 1) & 3;
	while (pad--)
		bputc(idx, 0);
	index_header.fastfind_start = btell(idx);
	build_fflist(trie_root);
	TRACE("Fastfind buffer stored, %d items on %d vertices.", index_header.num_fastfinds,
		index_header.num_fastfind_vertices);
}

/* returns subtree size, for branching type decision */
static uns compute_vertex_size(struct trie_node *t, uns last_match)
{
	uns s = 1;	// for vertex end

	ASSERT(t);

	ASSERT(t->match_cnt || t->ff_cnt);
	if (t->ff_cnt)				// fastfind list
		s += 5;
	else if (t->match_cnt <= 0x3f)		// small match
		s += 5;
	else					// big match
		s += 9;

	if (t->exact_match)
		if (t->exact_match <= 0x1f)	// small exact match
			s += 1;
		else
			s += 5;			// big exact match

	if (last_match+1 < t->match_size)	// character skip
	if (t->match_size - last_match > 1)	// character skip
		s += (t->match_size - last_match - 2) / 15 + 1 + t->match_size-last_match-1;

	uns ss = 0, sc = 0;
	struct trie_node *son = t->son;

	for (; son; son = son->brother) {
		ss += compute_vertex_size(son, t->match_size);
		sc++;
	}
	t->son_cnt = sc;

	if (sc)		// inner vertex
		if (sc <= 0x1f && s + 3*sc + ss + 1 <= 0xffff)	// can use near branching
			s += 1 + 3*sc;
		else						// have to use far branching
			s += 2 + 5*sc;
	t->subtree_size  = s+ss;
	return s+ss;
}

static void store_vertex(struct trie_node *t, uns last_match)
{
	int son_cnt = 0;
	struct trie_node *son_buf[256];

	if (t->ff_cnt) {			// fastfind list
		bputc(idx, 0x80 | t->ff_cnt);
		bputl(idx, t->ff_offset);
	}

	if (last_match+1 < t->match_size) {	// character skip
		uns ch;
		for (ch = 0; ch < t->match_size-last_match-1; ch++) {
			if (ch % 15 == 0)
				bputc(idx, 0x10 | MIN(t->match_size-last_match-1-ch, 15));
			bputc(idx, wordref_buf[t->wref].pd->phrase->text[wordref_buf[t->wref].wstart
					+ last_match + 1 + ch]);
		}
	}

	if (!t->ff_cnt)
		if (t->match_cnt <= 0x3f) {		// small match
			bputc(idx, 0x40 | t->match_cnt);
			bputl(idx, t->wref);
		}
		else {					// big match
			bputc(idx, 0x40);
			bputl(idx, t->wref);
			bputl(idx, t->match_cnt);
		}

	if (t->exact_match)
		if (t->exact_match <= 0x1f)			// small exact match
			bputc(idx, 0xc0 | t->exact_match);
		else {						// big exact match
			bputc(idx, 0xc0);
			bputl(idx, t->exact_match);
		}

	if (t->son_cnt)		// inner vertex
		if (t->son_cnt <= 0x1f && t->subtree_size <= 0xffff) {	// near branching
			bputc(idx, 0x20 | t->son_cnt);

			struct trie_node *son;
			uns ss = 0;
			for (son = t->son; son; son = son->brother) {
				son_buf[son_cnt++] = son;
				ss += son->subtree_size;
			}

			uns vsize = t->subtree_size - ss;		// compute vertex_size
			int bp = 0;
			u16 ra_buf[256];
			byte ch_buf[256];
			for (son = t->son; son; son = son->brother) {
				ch_buf[bp] = wordref_buf[son->wref].pd->phrase->text[wordref_buf[son->wref].wstart + t->match_size];
				ss -= son->subtree_size;
				ra_buf[bp++] = vsize + ss;
			}
			while (--bp >= 0) {
				bputc(idx, ch_buf[bp]);
				bputw(idx, ra_buf[bp]);
			}
		} else {						// far branching
			bputc(idx, 0x20);
			bputc(idx, t->son_cnt);

			struct trie_node *son;
			uns cpos = btell(idx) - index_header.trie_start + 5*t->son_cnt + 1;
			int bp = 0;
			byte ch_buf[256];
			u32 aa_buf[256];
			uns ss = 0;
			for (son = t->son; son; son = son->brother) {
				son_buf[son_cnt++] = son;
				ss += son->subtree_size;
			}

			for (son = t->son; son; son = son->brother) {
				ch_buf[bp] = wordref_buf[son->wref].pd->phrase->text[wordref_buf[son->wref].wstart + t->match_size];

				ss -= son->subtree_size;
				aa_buf[bp++] = cpos + ss;
			}
			while (--bp >= 0) {
				bputc(idx, ch_buf[bp]);
				bputl(idx, aa_buf[bp]);
			}
		}
	bputc(idx, 0xff);		// vertex end

	while (--son_cnt >= 0)
		store_vertex(son_buf[son_cnt], t->match_size);
}

static void store_trie(void)
{
	TRACE("Storing trie...");
	compute_vertex_size(trie_root, 0);
	index_header.trie_start = btell(idx);
	store_vertex(trie_root, 0);
	TRACE("Trie stored.");
}

static void build_trie(void)
{
	TRACE("Building trie...");
	trie_init();

	uns wr = 0;
	for (uns i=0; i<wordref_cnt; i++) {
		/* Skip multiple word occurences */
		if (i<wordref_cnt-1 && wordref_buf[i+1].wlen == wordref_buf[i].wlen
			&& strncmp(wordref_buf[i+1].pd->phrase->text + wordref_buf[i+1].wstart,
				wordref_buf[i].pd->phrase->text + wordref_buf[i].wstart, wordref_buf[i].wlen) == 0)
			continue;
		trie_insert(wr, i-wr+1);
		wr = i+1;
	}
	if (trie_root)
		trie_compute_matches(trie_root);
	TRACE("Trie built.");
}

uns kibitz_make_index(struct fastbuf *input, struct fastbuf *idxfb)
{
	idx = idxfb;
	bzero(&index_header, sizeof(index_header));
	phrase_list_mp = mp_new(1024*1024);
	phrase_cnt = 0;
	wordref_cnt = 0;
	wordref_buf_size = config.default_wordref_buf_size;

	write_tmp_header();
	load_phrases(input);
	store_phrasebuf();
	build_wordrefbuf();
	store_wordoffbuf();

	build_trie();

	store_fastfindbuf();
	store_trie();

#ifdef LOCAL_DEBUG
	trie_dump(trie_root, 0);
#endif

	mp_delete(phrase_list_mp);
	xfree(wordref_buf);
	trie_cleanup();

	/* Final header */
	brewind(idx);
	index_header.magic = INDEX_MAGIC;
	index_header.version = INDEX_VERSION;
	index_header.timestamp = time(NULL);
	index_header.index_size = bfilesize(idx);
	bwrite(idx, &index_header, sizeof(index_header));

	return phrase_cnt;
}

uns kibitz_make_index_file(byte *inputfile, byte *indexname)
{
	TRACE("Building index '%s' from '%s'", indexname, inputfile);
	struct fastbuf *input = bopen(inputfile, O_RDONLY, 65536);
	if (!input)
		return 0;
	struct fastbuf *idxfb = bopen(indexname, O_WRONLY | O_CREAT | O_TRUNC, 65536);
	if (!idxfb) {
		bclose(input);
		return 0;
	}
	uns cnt = kibitz_make_index(input, idxfb);
	bclose(input);
	bclose(idxfb);
	TRACE("Index built");
	return cnt;
}

